import os
from tqdm.notebook import tqdm
import matplotlib.pyplot as plt
from collections import OrderedDict
from gpu import *
from utils import *
import time


class BaseTrainer():
    def __init__(self, n_epochs=10, init_lr=0.01, file_name=None):
        """
        Base class for Trainer

        Parameters
        ----------
        n_epochs (int) : number of epochs. Default: 10

        init_lr (int) : initial learning rate. Default: 0.01

        """
        self.n_epochs = n_epochs
        self.init_lr = init_lr
        self.metric = OrderedDict()
        self.train_loss_list = np.array([])
        self.train_loss_list_iter = np.array([])
        self.valid_loss_list = np.array([])
        self.valid_loss_list_iter = np.array([])
        self.test_loss_list = np.array([])
        self.test_loss_list_iter = np.array([])
        self.best_valid_loss = 100
        self.file_name = file_name
        self.train_time = np.array([])

    def __repr__(self):
        return "Trainer"

    def get_train_loss_list(self):
        return self.train_loss_list

    def get_train_loss_list_iter(self):
        return self.train_loss_list_iter

    def get_valid_loss_list(self):
        return self.valid_loss_list

    def get_train_loss_list_iter(self):
        return self.valid_loss_list_iter


class EmbeddingTrainer(BaseTrainer):
    def __init__(self, model, optimizer, n_epochs=10, file_name=None):
        """
        Trainer for Classification

        Parameters
        ----------
        model : model for classification task

        critic : loss function class

        optimizer : optimizer for model optimizing

        n_epochs (int) : number of epochs. Default: 10

        init_lr (int) : initial learning rate. Default: 0.01

        """
        super().__init__(n_epochs, file_name=file_name)
        self.model = model
        self.optimizer = optimizer

        self.train_accuracy_list = np.array([0])
        self.valid_accuracy_list = np.array([0])

    def _forward(self, x, y):
        loss = self.model.forward(x, y)

        return loss

    def _backward(self):
        self.model.backward()

    def _update(self):
        self.optimizer.update(self.model)

    def train(self, train_dataloader, max_grad=None):
        """
        Train model with train/valid data

        Parameters
        ----------
        train_dataloader (iterable) : Train dataloader that can iterate with batch size in train dataset

        valid_dataloader (iterable) : Valid dataloader that can iterate with batch size in valid dataset

        """
        for epoch in range(self.n_epochs):
            start_time = time.time()
            print(f"epoch {epoch + 1}")

            tmp_train_loss = np.array([])
            iter_tmp_train_loss = np.array([])
            count = 0
            iter_num = len(train_dataloader)

            for x, y in tqdm(train_dataloader):
                count += 1
                train_loss = self._forward(x, y)
                self._backward()
                if max_grad is not None:
                    self.clip_grad(max_grad)
                self._update()

                tmp_train_loss = np.append(tmp_train_loss, train_loss)
                iter_tmp_train_loss = np.append(iter_tmp_train_loss, train_loss)

                print(f"train loss : {train_loss:.6f}    iter : {count}/{iter_num}\r", end="")

            epoch_train_loss = np.sum(tmp_train_loss) / len(tmp_train_loss)
            self.train_loss_list = np.append(self.train_loss_list, epoch_train_loss)

            print()
            print(f"epoch {epoch + 1} -- train loss : {epoch_train_loss}")

            finish_time = time.time()
            epoch_time = finish_time - start_time
            self.train_time = np.append(self.train_time, epoch_time)
            print(f"{epoch_time:.1f}s elapsed")

    def clip_grad(self, max_norm):
        total_norm = 0
        param_grad_dict = self.optimizer.param_grad_dict

        for layer_name in self.model.sequence:
            layer = self.model.network[layer_name]

            # only update differentiable layer
            if layer.differentiable:
                grad = layer.get_gradient()
                param_list = list(layer.parameter.keys())
                for param in param_list:
                    total_norm += np.sum(grad[param_grad_dict[param]] ** 2)

        total_norm = np.sqrt(total_norm)
        rate = max_norm / (total_norm + 1e-6)

        if rate < 1:
            for layer_name in self.model.sequence:
                layer = self.model.network[layer_name]

                # only update differentiable layer
                if layer.differentiable:
                    grad = layer.get_gradient()
                    param_list = list(layer.parameter.keys())
                    for param in param_list:
                        grad[param_grad_dict[param]] *= rate

    def get_train_accuracy_list(self):
        return self.train_accuracy_list

    def get_valid_accuracy_list(self):
        return self.valid_accuracy_list


class ClassificationTrainer(BaseTrainer):
    def __init__(self, model, critic, optimizer, n_epochs=10, init_lr=0.01, file_name=None):
        """
        Trainer for Classification

        Parameters
        ----------
        model : model for classification task

        critic : loss function class

        optimizer : optimizer for model optimizing

        n_epochs (int) : number of epochs. Default: 10

        init_lr (int) : initial learning rate. Default: 0.01

        """
        super().__init__(n_epochs, init_lr, file_name=file_name)
        self.model = model
        self.critic = critic
        self.optimizer = optimizer

        self.train_accuracy_list = np.array([0])
        self.valid_accuracy_list = np.array([0])
        self.test_accuracy_list = np.array([0])

    def _forward(self, x, y):
        pred = self.model(x)
        loss = self.critic(pred, y)

        return loss

    def _backward(self):
        self.model.backward(self.critic)

    def _update(self):
        self.optimizer.update(self.model)

    def train(self, train_dataloader, valid_dataloader=None, show_iter_num=1):
        """
        Train model with train/valid data

        Parameters
        ----------
        train_dataloader (iterable) : Train dataloader that can iterate with batch size in train dataset

        valid_dataloader (iterable) : Valid dataloader that can iterate with batch size in valid dataset

        """

        for epoch in range(self.n_epochs):
            start_time = time.time()
            print(f"epoch {epoch + 1}")
            self.model.train_state()
            tmp_train_loss = np.array([])
            iter_tmp_train_loss = np.array([])
            count = 0
            iter_num = len(train_dataloader)
            train_correct_num = 0
            y_num = 0
            for x, y in tqdm(train_dataloader):
                count += 1
                train_loss = self._forward(x, y)
                self._backward()
                self._update()

                tmp_train_loss = np.append(tmp_train_loss, train_loss)
                iter_tmp_train_loss = np.append(iter_tmp_train_loss, train_loss)

                if count % show_iter_num == 0:
                    iter_train_loss = np.sum(iter_tmp_train_loss) / show_iter_num
                    iter_tmp_train_loss = np.array([])
                    self.train_loss_list_iter = np.append(self.train_loss_list_iter, iter_train_loss)

                pred = np.argmax(self.critic.pred, axis=1)

                if y.ndim != 1:
                    y = np.argmax(y, axis=1)

                train_correct_num += np.sum(pred == y)
                y_num += len(y)
                train_temp_accuracy = train_correct_num / y_num * 100

                print(
                    f"train loss : {train_loss:.6f}    train accuarcy : {train_temp_accuracy:.6f}    iter : {count}/{iter_num}\r",
                    end="")

            epoch_train_loss = np.sum(tmp_train_loss) / len(tmp_train_loss)
            self.train_loss_list = np.append(self.train_loss_list, epoch_train_loss)

            dataset_len = train_dataloader.dataset_len()
            epoch_train_accuracy = train_correct_num / float(dataset_len) * 100
            self.train_accuracy_list = np.append(self.train_accuracy_list, epoch_train_accuracy)

            print()
            print(f"epoch {epoch + 1} -- train loss : {epoch_train_loss}    train accuarcy : {epoch_train_accuracy}")

            finish_time = time.time()
            epoch_time = finish_time - start_time
            self.train_time = np.append(self.train_time, epoch_time)
            print(f"{epoch_time:.1f}s elapsed")
            if valid_dataloader is not None:
                self._validate(valid_dataloader, epoch, show_iter_num=show_iter_num)
            else:
                print("----------------------------------------------------------------")

    def _validate(self, valid_dataloader, epoch, show_iter_num=1):
        self.model.eval_state()
        self.model.reset_rnn_state()
        tmp_valid_loss = np.array([])
        iter_tmp_valid_loss = np.array([])
        count = 0
        valid_correct_num = 0
        y_num = 0
        for x, y in tqdm(valid_dataloader):
            count += 1
            valid_loss = self._forward(x, y)
            tmp_valid_loss = np.append(tmp_valid_loss, valid_loss)
            pred = np.argmax(self.critic.pred, axis=1)
            if y.ndim != 1:
                y = np.argmax(y, axis=1)
            valid_correct_num += np.sum(pred == y)
            y_num += len(y)

            valid_temp_accuracy = valid_correct_num / y_num * 100
            iter_tmp_valid_loss = np.append(iter_tmp_valid_loss, valid_loss)

            if count % show_iter_num == 0:
                iter_valid_loss = np.sum(iter_tmp_valid_loss) / show_iter_num
                iter_tmp_valid_loss = np.array([])
                self.valid_loss_list_iter = np.append(self.valid_loss_list_iter, iter_valid_loss)

            print(f"valid loss : {valid_loss:.6f}    valid accuarcy : {valid_temp_accuracy:.6f}\r", end="")

        dataset_len = valid_dataloader.dataset_len()

        epoch_valid_loss = np.sum(tmp_valid_loss) / len(tmp_valid_loss)
        self.valid_loss_list = np.append(self.valid_loss_list, epoch_valid_loss)

        epoch_valid_accuracy = valid_correct_num / float(dataset_len) * 100
        self.valid_accuracy_list = np.append(self.valid_accuracy_list, epoch_valid_accuracy)
        self.model.reset_rnn_state()

        if epoch_valid_loss < self.best_valid_loss and self.file_name != None:
            self.model.save_params(self.file_name)
            self.best_valid_loss = epoch_valid_loss

        print()
        print(f"epoch {epoch + 1} -- valid loss : {epoch_valid_loss}    valid accuarcy : {epoch_valid_accuracy}")
        print("----------------------------------------------------------------")

    def show_error_graph(self, show_iter=False, valid=False):
        """
        Visualize training/validation error
        """
        if show_iter:

            if valid:
                plt.plot(to_cpu(self.valid_loss_list_iter), "-b", label='valid loss')
                plt.title("train/valid loss per iteration")
            else:
                plt.title("train loss per iteration")

            plt.plot(to_cpu(self.train_loss_list_iter), "-r", label='train loss')
            plt.legend(loc="upper right")
            plt.grid()
            plt.show()

        else:
            if valid:
                plt.plot(to_cpu(self.valid_loss_list), "-b", label='valid loss')
                plt.title("train/valid loss per epoch")
            else:
                plt.title("train loss per epoch")

            plt.plot(to_cpu(self.train_loss_list), "-r", label='train loss')

            plt.legend(loc="upper right")

            plt.grid()
            plt.show()

    def show_accuracy_graph(self, valid=False):
        """
        Visualize training/validation accuracy
        """
        if valid:
            plt.plot(to_cpu(self.valid_accuracy_list), "-b", label='valid accuracy')
            plt.title("train/valid accuracy per epoch")
        else:
            plt.title("train accuracy per epoch")

        plt.plot(to_cpu(self.train_accuracy_list), "-r", label='train accuracy')

        plt.legend(loc="lower right")

        plt.grid()
        plt.show()

    def add_metric(self, **metric):
        print(f"metic {metric.keys()} added")
        self.metric = dict(self.metric, metric)

    def eval_accuracy(self, test_dataloader, show_iter_num=1):
        self.model.eval_state()
        self.model.reset_rnn_state()
        tmp_test_loss = np.array([])
        iter_tmp_test_loss = np.array([])
        count = 0
        test_correct_num = 0
        y_num = 0
        for x, y in tqdm(test_dataloader):
            count += 1
            test_loss = self._forward(x, y)
            tmp_test_loss = np.append(tmp_test_loss, test_loss)
            pred = np.argmax(self.critic.pred, axis=1)
            if y.ndim != 1:
                y = np.argmax(y, axis=1)
            test_correct_num += np.sum(pred == y)
            y_num += len(y)

            iter_tmp_test_loss = np.append(iter_tmp_test_loss, test_loss)

            if count % show_iter_num == 0:
                iter_test_loss = np.sum(iter_tmp_test_loss) / show_iter_num
                iter_tmp_test_loss = np.array([])
                self.test_loss_list_iter = np.append(self.test_loss_list_iter, iter_test_loss)

        dataset_len = test_dataloader.dataset_len()

        epoch_test_loss = np.sum(tmp_test_loss) / len(tmp_test_loss)
        self.test_loss_list = np.append(self.test_loss_list, epoch_test_loss)

        epoch_test_accuracy = test_correct_num / float(dataset_len) * 100
        self.test_accuracy_list = np.append(self.test_accuracy_list, epoch_test_accuracy)
        self.model.reset_rnn_state()

        print()
        print(f"test loss : {epoch_test_loss}    test accuarcy : {epoch_test_accuracy}")
        print("----------------------------------------------------------------")

    def get_train_accuracy_list(self):
        return self.train_accuracy_list

    def get_valid_accuracy_list(self):
        return self.valid_accuracy_list


class LanguageModelTrainer(BaseTrainer):
    def __init__(self, model, critic, optimizer, n_epochs=10, init_lr=0.01, file_name=None):
        """
        Trainer for Language modeling

        Parameters
        ----------
        model : model for classification task

        critic : loss function class

        optimizer : optimizer for model optimizing

        n_epochs (int) : number of epochs. Default: 10

        init_lr (int) : initial learning rate. Default: 0.01

        """
        super().__init__(n_epochs, init_lr, file_name=file_name)
        self.model = model
        self.critic = critic
        self.optimizer = optimizer
        self.train_perplexity_list_iter = np.array([])
        self.valid_perplexity_list_iter = np.array([])
        self.train_perplexity_list = np.array([])
        self.valid_perplexity_list = np.array([])

    def _forward(self, x, y):
        pred = self.model(x)
        loss = self.critic(pred, y)

        return loss

    def _backward(self):
        self.model.backward(self.critic)

    def _update(self):
        self.optimizer.update(self.model)

    def train(self, train_dataloader, valid_dataloader=None, max_grad=None, show_iter_num=1):
        """
        Train model with train/valid data

        Parameters
        ----------
        train_dataloader (iterable) : Train dataloader that can iterate with batch size in train dataset

        valid_dataloader (iterable) : Valid dataloader that can iterate with batch size in valid dataset

        """
        for epoch in range(self.n_epochs):
            start_time = time.time()
            self.model.train_state()
            print(f"epoch {epoch + 1}")
            tmp_train_loss = np.array([])
            iter_tmp_train_loss = np.array([])
            tmp_train_perplexity = np.array([])
            iter_tmp_train_perplexity = np.array([])

            count = 0
            iter_num = len(train_dataloader)

            for x, y in tqdm(train_dataloader):
                count += 1
                # train loss
                train_loss = self._forward(x, y)
                self._backward()
                if max_grad is not None:
                    self.clip_grad(max_grad)
                self._update()
                tmp_train_loss = np.append(tmp_train_loss, train_loss)
                # train perplexity
                train_perplexity = np.exp(train_loss)
                tmp_train_perplexity = np.append(tmp_train_perplexity, train_perplexity)
                iter_tmp_train_perplexity = np.append(iter_tmp_train_perplexity, train_perplexity)
                iter_tmp_train_loss = np.append(iter_tmp_train_loss, train_loss)

                if count % show_iter_num == 0:
                    iter_train_loss = np.sum(iter_tmp_train_loss) / show_iter_num
                    iter_tmp_train_loss = np.array([])
                    self.train_loss_list_iter = np.append(self.train_loss_list_iter, iter_train_loss)

                    iter_train_perplexity = np.sum(iter_tmp_train_perplexity) / show_iter_num
                    iter_tmp_train_perplexity = np.array([])
                    self.train_perplexity_list_iter = np.append(self.train_perplexity_list_iter, iter_train_perplexity)

                print(
                    f"train loss : {train_loss:.6f}    train perplexity : {train_perplexity:.6f}    iter : {count}/{iter_num}\r",
                    end="")

            epoch_train_loss = np.sum(tmp_train_loss) / count
            self.train_loss_list = np.append(self.train_loss_list, epoch_train_loss)

            epoch_train_perplexity = np.sum(tmp_train_perplexity) / count
            self.train_perplexity_list = np.append(self.train_perplexity_list, epoch_train_perplexity)

            print()
            print(
                f"epoch {epoch + 1} -- train loss : {epoch_train_loss:.6f}    train perplexity : {epoch_train_perplexity:.6f}")
            finish_time = time.time()
            epoch_time = finish_time - start_time
            self.train_time = np.append(self.train_time, epoch_time)
            print(f"{epoch_time:.1f}s elapsed")

            if valid_dataloader is not None:
                self._validate(valid_dataloader, epoch, show_iter_num=show_iter_num)
            else:
                print("----------------------------------------------------------------")
                print()

    def _validate(self, valid_dataloader, epoch, show_iter_num=1):
        self.model.eval_state()
        self.model.reset_rnn_state()
        tmp_valid_loss = np.array([])
        iter_tmp_valid_loss = np.array([])
        tmp_valid_perplexity = np.array([])
        iter_tmp_valid_perplexity = np.array([])
        count = 0
        for x, y in tqdm(valid_dataloader):
            count += 1
            # valid loss
            valid_loss = self._forward(x, y)
            tmp_valid_loss = np.append(tmp_valid_loss, valid_loss)
            # valid perplexity
            valid_perplexity = np.exp(valid_loss)
            tmp_valid_perplexity = np.append(tmp_valid_perplexity, valid_perplexity)
            iter_tmp_valid_perplexity = np.append(iter_tmp_valid_perplexity, valid_perplexity)

            if count % show_iter_num == 0:
                iter_valid_loss = np.sum(iter_tmp_valid_loss) / show_iter_num
                iter_tmp_valid_loss = np.array([])
                self.valid_loss_list_iter = np.append(self.valid_loss_list_iter, iter_valid_loss)

                iter_valid_perplexity = np.sum(iter_tmp_valid_perplexity) / show_iter_num
                iter_tmp_valid_perplexity = np.array([])
                self.valid_perplexity_list_iter = np.append(self.valid_perplexity_list_iter, iter_valid_perplexity)

            print(f"valid loss : {valid_loss:.6f}    valid perplexity : {valid_perplexity:.6f}\r", end="")

        epoch_valid_loss = np.sum(tmp_valid_loss) / count
        self.valid_loss_list = np.append(self.valid_loss_list, epoch_valid_loss)

        epoch_valid_perplexity = np.sum(tmp_valid_perplexity) / count
        self.valid_perplexity_list = np.append(self.valid_perplexity_list, epoch_valid_perplexity)
        self.model.reset_rnn_state()

        if epoch_valid_loss < self.best_valid_loss and self.file_name != None:
            self.model.save_params(self.file_name)
            self.best_valid_loss = epoch_valid_loss

        print()
        print(
            f"epoch {epoch + 1} -- valid loss : {epoch_valid_loss:.6f}    valid perplexity : {epoch_valid_perplexity:.6f}")
        print("----------------------------------------------------------------")
        print()

    def clip_grad(self, max_norm):
        total_norm = 0
        param_grad_dict = self.optimizer.param_grad_dict

        for layer_name in self.model.sequence:
            layer = self.model.network[layer_name]

            # only update differentiable layer
            if layer.differentiable:
                grad = layer.get_gradient()
                param_list = list(layer.parameter.keys())
                for param in param_list:
                    total_norm += np.sum(grad[param_grad_dict[param]] ** 2)

        total_norm = np.sqrt(total_norm)
        rate = max_norm / (total_norm + 1e-6)

        if rate < 1:
            for layer_name in self.model.sequence:
                layer = self.model.network[layer_name]

                # only update differentiable layer
                if layer.differentiable:
                    grad = layer.get_gradient()
                    param_list = list(layer.parameter.keys())
                    for param in param_list:
                        grad[param_grad_dict[param]] *= rate

    def show_error_graph(self, show_iter=False, valid=False):
        """
        Visualize training/validation error
        """
        if show_iter:

            if valid:
                plt.plot(to_cpu(self.valid_loss_list_iter), "-b", label='valid loss')
                plt.title("train/valid loss per iteration")
            else:
                plt.title("train loss per iteration")

            plt.plot(to_cpu(self.train_loss_list_iter), "-r", label='train loss')
            plt.legend(loc="upper right")
            plt.grid()
            plt.show()

        else:
            if valid:
                plt.plot(to_cpu(self.valid_loss_list), "-b", label='valid loss')
                plt.title("train/valid loss per epoch")
            else:
                plt.title("train loss per epoch")

            plt.plot(to_cpu(self.train_loss_list), "-r", label='train loss')

            plt.legend(loc="upper right")

            plt.grid()
            plt.show()

    def show_perplexity_graph(self, show_iter=False, valid=False):
        """
        Visualize training/validation perplexity
        """
        plt.ylim(-1, 401)
        if show_iter:

            if valid:
                plt.plot(to_cpu(self.valid_perplexity_list_iter), "-b", label='valid perplexity')
                plt.title("train/valid perplexity per iteration")
            else:
                plt.title("train perplexity per iteration")

            plt.plot(to_cpu(self.train_perplexity_list_iter), "-r", label='train perplexity')
            plt.legend(loc="upper right")
            plt.grid()
            plt.show()

        else:
            if valid:
                plt.plot(to_cpu(self.valid_perplexity_list), "-b", label='valid perplexity')
                plt.title("train/valid perplexity per epoch")
            else:
                plt.title("train perplexity per epoch")

            plt.plot(to_cpu(self.train_perplexity_list), "-r", label='train perplexity')

            plt.legend(loc="upper right")

            plt.grid()
            plt.show()

    def eval_perplexity(self, test_dataloader):
        print('Test perplexity ...')
        self.model.eval_state()
        self.model.reset_rnn_state()

        tmp_test_loss = np.array([])
        tmp_test_perplexity = np.array([])

        count = 0
        for x, y in tqdm(test_dataloader):
            count += 1
            # test loss
            test_loss = self._forward(x, y)
            tmp_test_loss = np.append(tmp_test_loss, test_loss)
            # test perplexity
            test_perplexity = np.exp(test_loss)
            tmp_test_perplexity = np.append(tmp_test_perplexity, test_perplexity)

        epoch_test_loss = np.sum(tmp_test_loss) / count
        epoch_test_perplexity = np.sum(tmp_test_perplexity) / count

        print()
        print(f"test loss : {epoch_test_loss:.6f}    test perplexity : {epoch_test_perplexity:.6f}")

    def get_train_ppl_list(self):
        return self.train_perplexity_list

    def get_train_ppl_list_iter(self):
        return self.train_perplexity_list_iter

    def get_valid_ppl_list(self):
        return self.valid_perplexity_list

    def get_train_ppl_list_iter(self):
        return self.valid_perplexity_list_iter


class Seq2SeqTrainer(BaseTrainer):
    def __init__(self, model, critic, optimizer, vocab, n_epochs=10, init_lr=0.01, reverse=False, file_name=None):
        """
        Trainer for Classification

        Parameters
        ----------
        model : model for classification task

        critic : loss function class

        optimizer : optimizer for model optimizing

        n_epochs (int) : number of epochs. Default: 10

        init_lr (int) : initial learning rate. Default: 0.01

        """
        super().__init__(n_epochs, init_lr, file_name=file_name)
        self.model = model
        self.critic = critic
        self.optimizer = optimizer
        self.vocab = vocab
        self.verbose = 0
        self.is_reverse = reverse

        self.train_accuracy_list = np.array([0])
        self.valid_accuracy_list = np.array([0])
        self.test_accuracy_list = np.array([0])

    def _forward(self, x, y):
        decoder_x, decoder_y = y[:, :-1], y[:, 1:]
        pred = self.model(x, decoder_x)
        loss = self.critic(pred, decoder_y)

        return loss

    def _backward(self):
        self.model.backward(self.critic)

    def _update(self):
        self.optimizer.update(self.model)

    def train(self, train_dataloader, valid_x, valid_y, max_grad=None, show_iter_num=1, verbose=10):
        """
        Train model with train/valid data

        Parameters
        ----------
        train_dataloader (iterable) : Train dataloader that can iterate with batch size in train dataset

        valid_dataloader (iterable) : Valid dataloader that can iterate with batch size in valid dataset

        """
        for epoch in range(self.n_epochs):
            self.verbose = verbose
            start_time = time.time()
            print(f"epoch {epoch + 1}")
            self.model.train_state()

            tmp_train_loss = np.array([])
            iter_tmp_train_loss = np.array([])
            count = 0
            iter_num = len(train_dataloader)
            train_correct_num = 0
            y_num = 0

            for x, y in tqdm(train_dataloader):
                count += 1
                train_loss = self._forward(x, y)
                self._backward()
                if max_grad is not None:
                    self.clip_grad(max_grad)
                self._update()

                tmp_train_loss = np.append(tmp_train_loss, train_loss)
                iter_tmp_train_loss = np.append(iter_tmp_train_loss, train_loss)

                print(f"train loss : {train_loss:.6f}    iter : {count}/{iter_num}\r", end="")

            epoch_train_loss = np.sum(tmp_train_loss) / len(tmp_train_loss)
            self.train_loss_list = np.append(self.train_loss_list, epoch_train_loss)

            print()
            print(f"epoch {epoch + 1} -- train loss : {epoch_train_loss}")

            finish_time = time.time()
            epoch_time = finish_time - start_time
            self.train_time = np.append(self.train_time, epoch_time)
            print(f"{epoch_time:.1f}s elapsed")

            self.model.eval_state()
            correct_num = 0
            for i in range(len(valid_x)):
                question, correct = valid_x[[i]], valid_y[[i]]
                correct_num += self.eval_seq2seq(question, correct)

                print(f"valid accuarcy : {correct_num / len(valid_x):.6f}\r", end="")

            epoch_valid_accuracy = (correct_num / float(len(valid_x))) * 100
            self.valid_accuracy_list = np.append(self.valid_accuracy_list, epoch_valid_accuracy)

            print()
            print(f"epoch {epoch + 1} -- valid accuarcy : {epoch_valid_accuracy}")
            print("----------------------------------------------------------------")

    def clip_grad(self, max_norm):
        total_norm = 0
        param_grad_dict = self.optimizer.param_grad_dict

        for layer_name in self.model.sequence:
            layer = self.model.network[layer_name]

            # only update differentiable layer
            if layer.differentiable:
                grad = layer.get_gradient()
                param_list = list(layer.parameter.keys())
                for param in param_list:
                    total_norm += np.sum(grad[param_grad_dict[param]] ** 2)

        total_norm = np.sqrt(total_norm)
        rate = max_norm / (total_norm + 1e-6)

        if rate < 1:
            for layer_name in self.model.sequence:
                layer = self.model.network[layer_name]

                # only update differentiable layer
                if layer.differentiable:
                    grad = layer.get_gradient()
                    param_list = list(layer.parameter.keys())
                    for param in param_list:
                        grad[param_grad_dict[param]] *= rate

    def eval_seq2seq(self, src, tgt):
        tgt = tgt.flatten()
        start_id = tgt[0]
        tgt = tgt[1:]
        pred = self.model.generate(src, start_id, len(tgt))

        src = ''.join([self.vocab[int(c)] for c in src.flatten()])
        tgt = ''.join([self.vocab[int(c)] for c in tgt])
        pred = ''.join([self.vocab[int(c)] for c in pred])

        if self.verbose > 0:
            if self.is_reverse:
                src = src[::-1]

            colors = {'ok': '\033[92m', 'fail': '\033[91m', 'close': '\033[0m'}
            print('Q', src)
            print('T', tgt)

            is_windows = os.name == 'nt'

            if pred == tgt:
                mark = colors['ok'] + '☑' + colors['close']
                if is_windows:
                    mark = 'O'
                print(mark + ' ' + pred)
            else:
                mark = colors['fail'] + '☒' + colors['close']
                if is_windows:
                    mark = 'X'
                print(mark + ' ' + pred)
            print('---')

            self.verbose -= 1

        return 1 if pred == tgt else 0

    def get_train_accuracy_list(self):
        return self.train_accuracy_list

    def get_valid_accuracy_list(self):
        return self.valid_accuracy_list


from collections import Counter
from gpu import *
from function import *
from model import *


class BaseFunction():
    def __init__(self):
        self.differentiable = False
        self.changeability = False
        self.mixed_precision = False
        self.tied = False

    def __call__(self, arg):
        result = self._forward(arg)
        return result

    def _forward(self, x):
        pass

    def __repr__(self) -> str:
        return "Function"

    def _mixed_precision_training(self):
        self.mixed_precision = True


class Step(BaseFunction):
    def __init__(self):
        super().__init__()

    def _forward(self, x):
        return step_function(x)


class Identity(BaseFunction):
    def __init__(self):
        super().__init__()

    def _forward(self, x):
        return identity(x)


class Sigmoid(BaseFunction):
    def __init__(self):
        super().__init__()
        self.differentiable = False
        self.out = None

    def _forward(self, x):
        out = sigmoid(x)
        self.out = out

        return out

    def _backward(self, input):
        result = input * self.out * (1.0 - self.out)

        return result


class Tanh(BaseFunction):
    def __init__(self):
        super().__init__()
        self.differentiable = False
        self.out = None

    def _forward(self, x):
        out = np.tanh(x)
        self.out = out

        return out

    def _backward(self, input):
        result = input * (1 - self.out ** 2)

        return result


class Softmax(BaseFunction):
    def __init__(self):
        super().__init__()
        self.out = None

    def _forward(self, x):
        self.out = softmax(x)
        return self.out

    def _backward(self, input):
        dx = self.out * input
        sumdx = np.sum(dx, axis=1, keepdims=True)
        dx -= self.out * sumdx
        return dx


class ReLU(BaseFunction):
    def __init__(self):
        self.differentiable = False
        self.mask = None
        super().__init__()

    def _forward(self, x):
        self.mask = (x <= 0)
        result = x.copy()
        result[self.mask] = 0

        return result

    def _backward(self, input):
        input[self.mask] = 0
        result = input

        return result


class CrossEntropyLoss():
    def __init__(self, loss_scaling=None):
        self.loss = None
        self.pred = None
        self.true = None
        self.cache = None
        self.ignore_label = -1
        self.mixed_precision = False
        self.scaling_factor = loss_scaling

    def __repr__(self):
        return "Function"

    def __call__(self, *arg):
        result = self._forward(*arg)
        return result

    def _forward(self, pred, true):
        if pred.ndim == 3:
            batch_size, n_timestep, vocab_size = pred.shape
            if true.ndim == 3:
                true = true.argmax(axis=2)

            mask = (true != self.ignore_label)

            self.pred = pred.reshape(batch_size * n_timestep, vocab_size)
            self.true = true.reshape(batch_size * n_timestep)
            mask = mask.reshape(batch_size * n_timestep)
            pred_sentence = softmax(self.pred)
            ls = np.log(pred_sentence[np.arange(batch_size * n_timestep), self.true])
            ls *= mask
            self.loss = -np.sum(ls)
            self.loss /= mask.sum()
            self.cache = (self.true, pred_sentence, mask, (batch_size, n_timestep, vocab_size))

        else:
            self.pred = softmax(pred)
            self.true = true
            self.loss = cross_entropy_error(self.pred, self.true)

        return self.loss

    def _backward(self):
        if self.cache is not None:
            true, pred_sentence, mask, (batch_size, n_timestep, vocab_size) = self.cache

            dx = pred_sentence
            dx[np.arange(batch_size * n_timestep), true] -= 1
            dx *= 1
            dx /= mask.sum()
            dx *= mask[:, np.newaxis]  # ignore_label에 해당하는 데이터는 기울기를 0으로 설정

            dx = dx.reshape((batch_size, n_timestep, vocab_size))

            return dx

        else:
            batch_size = self.true.shape[0]
            if self.true.size == self.pred.size:
                dx = (self.pred - self.true) / batch_size
            else:
                dx = self.pred.copy()
                dx[np.arange(batch_size), self.true] -= 1
                dx = dx / batch_size

            if self.scaling_factor != None:
                dx = dx * self.scaling_factor

            return dx


class BinaryCrossEntropyLoss(BaseFunction):
    def __init__(self, loss_scaling=None):
        super().__init__()
        self.loss = None
        self.pred = None
        self.true = None
        self.mixed_precision = False
        self.scaling_factor = loss_scaling

    def __repr__(self) -> str:
        return "Function"

    def __call__(self, *arg):
        result = self._forward(*arg)
        return result

    def _forward(self, pred, true):
        self.pred = sigmoid(pred)
        self.true = true
        self.loss = cross_entropy_error(np.c_[1 - self.pred, self.pred], self.true)

        return self.loss

    def _backward(self, dout=1):
        batch_size = self.true.shape[0]

        dx = (self.pred - self.true) * dout / batch_size

        return dx


class NegativeSampler(BaseFunction):
    def __init__(self, corpus, power=0.75, sample_size=2):
        super().__init__()
        # power가 0에 가까울수록 출현 빈도 낮은 sample들이 많이 뽑히게 됨

        # 몇 개 추출할 것인지
        self.sample_size = sample_size
        self.vocab_size = None
        self.word_p = None

        counts = Counter()
        for word_id in corpus:
            counts[word_id] += 1

        vocab_size = len(counts)
        self.vocab_size = vocab_size

        self.word_p = np.zeros(vocab_size)
        for i in range(vocab_size):
            self.word_p[i] = counts[i]

        self.word_p = np.power(self.word_p, power)
        self.word_p /= np.sum(self.word_p)

    def __repr__(self) -> str:
        return "Sampler"

    def negative_sampling(self, target):
        batch_size = target.shape[0]

        if not GPU:
            negative_sample = np.zeros((batch_size, self.sample_size), dtype=np.int32)

            for i in range(batch_size):
                p = self.word_p.copy()
                target_idx = target[i]
                p[target_idx] = 0
                p /= p.sum()
                negative_sample[i, :] = np.random.choice(self.vocab_size, size=self.sample_size, replace=False, p=p)
        else:
            negative_sample = np.random.choice(self.vocab_size, size=(batch_size, self.sample_size), replace=True,
                                               p=self.word_p)

        return negative_sample





from gpu import *
import copy
from collections import OrderedDict


class BaseTokenizer():
    def __init__(self):
        self.word_to_id = OrderedDict()
        self.id_to_word = OrderedDict()
        self.punctuation_list = [".", ",", "?", "!"]
        self.special_token = OrderedDict()
        self.special_token["unk"] = "[UNK]"
        token_value = list(self.special_token.values())
        for i, token in enumerate(token_value):
            self.word_to_id[token], self.id_to_word[i] = i, token

        self.n_tokens = len(self.special_token.keys())

    def add_speical_token(self, name, token):
        self.special_token[name] = token

    def __repr__(self) -> str:
        return "Tokenizer"


class SpaceTokenizer(BaseTokenizer):
    def __init__(self):
        super().__init__()

    def train(self, text):
        """
        train vocab based on space

        Parameters
        ----------
        text (str or iterable) : sentence to be tokenized

        """
        if isinstance(text, str):
            text = text.lower()
            for punc in self.punctuation_list:
                text = text.replace(f"{punc}", f" {punc}")
            words = text.split(" ")

        else:
            text_copy = copy.deepcopy(text)
            words = []
            for index, sentence in enumerate(text_copy):
                sentence = sentence.lower()
                for punc in self.punctuation_list:
                    sentence = sentence.replace(f"{punc}", f" {punc}")
                words += sentence.split(" ")

        for i, word in enumerate(set(words)):
            self.word_to_id[word], self.id_to_word[i + self.n_tokens] = i + self.n_tokens, word

    def encode(self, text):
        """
        convert text into integer id
        """
        encoded_text = np.array([]).astype(np.int32)

        if isinstance(text, str):
            text = text.lower()
            for punc in self.punctuation_list:
                text = text.replace(f"{punc}", f" {punc}")
            words = text.split(" ")

            for word in words:
                try:
                    encoded_word = self.word_to_id[word]
                    encoded_text = np.append(encoded_text, encoded_word)
                except KeyError:
                    encoded_text = np.append(encoded_text, self.word_to_id["[UNK]"])


        else:
            text_copy = copy.deepcopy(text)

            for index, sentence in enumerate(text_copy):
                sentence = sentence.lower()
                for punc in self.punctuation_list:
                    sentence = sentence.replace(f"{punc}", f" {punc}")
                    sentence.split(" ")

                encoded_sentence = np.array([])
                for word in sentence:
                    try:
                        encoded_word = self.word_to_id[word]
                        encoded_sentence = np.append(encoded_sentence, encoded_word)
                    except KeyError:
                        encoded_sentence = np.append(encoded_sentence, self.word_to_id["[UNK]"])

                encoded_text = np.append(encoded_text, encoded_sentence)

        return encoded_text

    def decode(self, id_list: list) -> str:
        id_list_copy = copy.deepcopy(id_list)
        decoded_id = []
        if id_list_copy.ndim == 1:
            for index, id in enumerate(id_list_copy):
                temp = self.id_to_word[id]
                decoded_id.append(temp)

            id_sentence = " ".join(decoded_id)

            return id_sentence

        else:
            temp_decoded = []
            for id_sentence in id_list_copy:
                for index, id in enumerate(id_sentence):
                    temp_decoded.append(self.id_to_word[id])

                temp_decoded_str = " ".join(temp_decoded)
                decoded_id.append(temp_decoded_str)

            return decoded_id

    def get_vocab(self):
        return self.word_to_id

    def get_id(self):
        return self.id_to_word


def cos_similarity(x, y, eps=1e-8):
    '''코사인 유사도 산출

    :param x: 벡터
    :param y: 벡터
    :param eps: '0으로 나누기'를 방지하기 위한 작은 값
    :return:
    '''
    nx = x / (np.sqrt(np.sum(x ** 2)) + eps)
    ny = y / (np.sqrt(np.sum(y ** 2)) + eps)
    return np.dot(nx, ny)


def analogy(a, b, c, word_to_id, id_to_word, word_matrix, top=5, answer=None):
    for word in (a, b, c):
        if word not in word_to_id:
            print('%s(을)를 찾을 수 없습니다.' % word)
            return

    print('\n[analogy] ' + a + ':' + b + ' = ' + c + ':?')
    a_vec, b_vec, c_vec = word_matrix[word_to_id[a]], word_matrix[word_to_id[b]], word_matrix[word_to_id[c]]
    query_vec = b_vec - a_vec + c_vec
    query_vec = normalize(query_vec)

    similarity = np.dot(word_matrix, query_vec)

    if answer is not None:
        print("==>" + answer + ":" + str(np.dot(word_matrix[word_to_id[answer]], query_vec)))

    count = 0
    for i in (-1 * similarity).argsort():
        if np.isnan(similarity[i]):
            continue
        if id_to_word[i] in (a, b, c):
            continue
        print(' {0}: {1}'.format(id_to_word[i], similarity[i]))

        count += 1
        if count >= top:
            return


def normalize(x):
    if x.ndim == 2:
        s = np.sqrt((x * x).sum(1))
        x /= s.reshape((s.shape[0], 1))
    elif x.ndim == 1:
        s = np.sqrt((x * x).sum())
        x /= s
    return x


def most_similar(query, word_to_id, id_to_word, word_matrix, top=5):
    '''유사 단어 검색

    :param query: 쿼리(텍스트)
    :param word_to_id: 단어에서 단어 ID로 변환하는 딕셔너리
    :param id_to_word: 단어 ID에서 단어로 변환하는 딕셔너리
    :param word_matrix: 단어 벡터를 정리한 행렬. 각 행에 해당 단어 벡터가 저장되어 있다고 가정한다.
    :param top: 상위 몇 개까지 출력할 지 지정
    '''
    if query not in word_to_id:
        print('%s(을)를 찾을 수 없습니다.' % query)
        return

    print('\n[query] ' + query)
    query_id = word_to_id[query]
    query_vec = word_matrix[query_id]

    # 코사인 유사도 계산
    vocab_size = len(id_to_word)

    similarity = np.zeros(vocab_size)
    for i in range(vocab_size):
        similarity[i] = cos_similarity(word_matrix[i], query_vec)

    # 코사인 유사도를 기준으로 내림차순으로 출력
    count = 0
    for i in (-1 * similarity).argsort():
        if id_to_word[i] == query:
            continue
        print(' %s: %s' % (id_to_word[i], similarity[i]))

        count += 1
        if count >= top:
            return


def convert_one_hot(corpus, vocab_size):
    '''원핫 표현으로 변환

    :param corpus: 단어 ID 목록(1차원 또는 2차원 넘파이 배열)
    :param vocab_size: 어휘 수
    :return: 원핫 표현(2차원 또는 3차원 넘파이 배열)
    '''
    N = corpus.shape[0]

    if corpus.ndim == 1:
        one_hot = np.zeros((N, vocab_size), dtype=np.int32)
        for idx, word_id in enumerate(corpus):
            one_hot[idx, word_id] = 1

    elif corpus.ndim == 2:
        C = corpus.shape[1]
        one_hot = np.zeros((N, C, vocab_size), dtype=np.int32)
        for idx_0, word_ids in enumerate(corpus):
            for idx_1, word_id in enumerate(word_ids):
                one_hot[idx_0, idx_1, word_id] = 1

    return one_hot


def create_contexts_target(corpus, window_size=1):
    '''맥락과 타깃 생성

    :param corpus: 말뭉치(단어 ID 목록)
    :param window_size: 윈도우 크기(윈도우 크기가 1이면 타깃 단어 좌우 한 단어씩이 맥락에 포함)
    :return:
    '''
    target = corpus[window_size:-window_size]
    contexts = []

    for idx in range(window_size, len(corpus) - window_size):
        cs = []
        for t in range(-window_size, window_size + 1):
            if t == 0:
                continue
            cs.append(corpus[idx + t])
        contexts.append(cs)

    return np.array(contexts), np.array(target)